#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed by avariant - Copyright May 2009

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#Qhull Copyrigght
#-----------------------------------------------------------------------------
# Name		:   qhull.rb
# Original Date	:   May 5 2009 - version 1.0
# Description	:   Bridge and Management to the QHull libraries
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************

#*************************************************************************************************
#Qhull Copyright

#Qhull, Copyright (c) 1993-2003

#      The National Science and Technology Research Center for
#        Computation and Visualization of Geometric Structures
#                        (The Geometry Center)
#                       University of Minnesota

#***************************************************************************************************

module SU_Qhull

class Qhull

require 'sketchup.rb'

#initialize class by getting the qhull directory path
def self.initialize
	@qhull_path = Sketchup.find_support_file("Plugins/qhull")
	@qhull_path = @qhull_path + "/qhull-2003.1"
end

#Recursivelt convert the selection to vertices
def self.GetPoints(item, array, xform, box)
	entities = nil
	new_xform = Geom::Transformation.new(xform)
	
	if (item.class == Sketchup::ComponentInstance)
		entities = item.definition.entities
		new_xform = xform * item.transformation
	elsif (item.class == Sketchup::Group)
		entities = item.entities
		new_xform = xform * item.transformation
	elsif (item.class == Sketchup::Selection)
		item.each do |e|
			GetPoints(e, array, Geom::Transformation.new, box)
		end
	elsif (item.class == Sketchup::Edge)
		pt = item.start.position.transform(xform)
		array << pt
		box.add(pt)
		pt2 = item.end.position.transform(xform)
		array << pt2
		box.add(pt2)
	elsif (item.class==Sketchup::ConstructionPoint)
		pt = item.position.transform(xform)
		array << pt
		box.add(pt)
	end
	
	if not entities == nil
		entities.each do |e|
			GetPoints(e, array, new_xform, box)
		end
	end
end

#Common results file processing
def self.CommonResults(voronoi, delaunay)
	return if not (File.exists?("result.txt"))
	
	#read result file
	lines_array = IO.readlines("result.txt")
	count = 0
	dimension = 3
	vertex_count = 0
	vertex_array = []
	face_array = []
	dump_count =0
	#loop through each line, processing according to line number
	lines_array.each do |e|
		#Graph dimension
		if (count==0)
			dimension = e.split(' ')[0].to_i
		#vertex and region? count
		elsif (count==1)
			array = e.split(' ')
			vertex_count = array[0].to_i
			if (voronoi)
				dump_count = array[1].to_i
			end
		#3 value vertices
		elsif (count>1 and count <= (vertex_count + 1))
			array = e.split(' ')
			x = array[0].to_f
			y = array[1].to_f
			z = array[2].to_f
			pt = Geom::Point3d.new(x, y, z)
			vertex_array << pt
		#faces, may be voronoi style, or delaunay style
		elsif (count > (vertex_count + 1 + dump_count))
			array = e.split(' ')
			face = GetFace(array, voronoi)
			if not face==nil
				if (delaunay)
					face_array << [face[0], face[1], face[2]]
					face_array << [face[1], face[2], face[3]]
				else
					face_array << face
				end
			end
		end
		count = count + 1
	end
	#convert to geometry
	geometry(vertex_array, face_array)
end

#gets the face data from the array.  voronoi is processed slightly different
def self.GetFace(array, voronoi)
	return nil if array.length <=3
	face = []
	index_count = array[0].to_i
	return nil if array.length < (index_count + 1)
	start_index = 1
	if (voronoi)
		start_index = 3
	end
	
	for i in start_index..index_count do
		idx = array[i].to_i
		return nil if (voronoi and (idx==0))
		face << idx
	end
	return nil if (face.length<3)
	return face
end

#Create Voronoi model using data points fed from rbox
def self.Voronoi(count)
	Dir.chdir(@qhull_path)
	if not system("rbox " + count.to_s + " t B5 | qvoronoi o Fv TO result.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(true, false)
end

#Create Voronoi using data points from the current selection
def self.Selected_Voronoi(sphere, cube)
	Dir.chdir(@qhull_path)
	sel = Sketchup.active_model.selection
	return if (sel.empty?)
	array = []
	box = Geom::BoundingBox.new
	GetPoints(sel, array, Geom::Transformation.new, box)
	
	pta = box.corner(3)
	ptb = box.corner(4)
	diag = pta.vector_to(ptb)
	len = diag.length / 2
	len = len * 1.1
	center = box.center
	
	#adds an array of points in a sphere around the selection
	if sphere
		for i in 0..5 do
			az = Geom::Transformation.rotation(ORIGIN, Z_AXIS, (i*PI/3))
			for j in -2..2 do
				el = az * Geom::Transformation.rotation(ORIGIN, X_AXIS, (j * PI/6))
				sphere_pt = Geom::Point3d.new(0,len,0)
				sphere_pt = sphere_pt.transform(el)
				new_pt = center.offset(sphere_pt.vector_to(ORIGIN))
				array << sphere_pt
			end
		end
	end
	
	#adds an array of points in a cube around the selection
	if cube
		if not box.empty?
			for i in 0..7 do
				corner = box.corner(i)
				c_c = corner - center
				corner = corner.offset(c_c, c_c.length * 0.1)
				array << corner
			end
		end
	end
	
	file = File.new("input.txt", "w")
	file.puts("3 rbox " + array.length.to_s + " t")
	file.puts(array.length.to_s)
	array.each do |v|
		file.puts(v.x.to_f.to_s + " " + v.y.to_f.to_s + " " + v.z.to_f.to_s)
	end
	file.puts("")
	file.close()
	if not system("qvoronoi o Fv TO result.txt < input.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(true, false)
end

#Create Convex Hull model using data points fed from rbox
def self.Convex(count)
	Dir.chdir(@qhull_path)
	if not system("rbox " + count.to_s + " t B5 | qconvex o TO result.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, false)
end

#Create Convex Hull model using data points from the selection
def self.Selected_Convex()
	Dir.chdir(@qhull_path)
	sel = Sketchup.active_model.selection
	return if (sel.empty?)
	array = []
	GetPoints(sel, array, Geom::Transformation.new, Geom::BoundingBox.new)
	
	file = File.new("input.txt", "w")
	file.puts("3 rbox " + array.length.to_s + " t")
	file.puts(array.length.to_s)
	array.each do |v|
		file.puts(v.x.to_f.to_s + " " + v.y.to_f.to_s + " " + v.z.to_f.to_s)
	end
	file.puts("")
	file.close()
	if not system("qconvex o TO result.txt < input.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, false)
end

#Create Delaunay model using data points fed from rbox
def self.Delaunay(count)
	Dir.chdir(@qhull_path)
	if not system("rbox " + count.to_s + " t B5 | qdelaunay Qj o TO result.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, true)
end

#Create Delaunay model using data points from selection
def self.Selected_Delaunay()
	Dir.chdir(@qhull_path)
	sel = Sketchup.active_model.selection
	return if (sel.empty?)
	array = []
	GetPoints(sel, array, Geom::Transformation.new, Geom::BoundingBox.new)
	
	file = File.new("input.txt", "w")
	file.puts("3 rbox " + array.length.to_s + " t")
	file.puts(array.length.to_s)
	array.each do |v|
		file.puts(v.x.to_f.to_s + " " + v.y.to_f.to_s + " " + v.z.to_f.to_s)
	end
	file.puts("")
	file.close()
	if not system("qdelaunay Qj o TO result.txt < input.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, true)
end

#Create Half point model using data points fed from rbox
def self.Half(count)
	Dir.chdir(@qhull_path)
	if not system("rbox " + count.to_s + " t B5 | qconvex FQ FV n | qhalf Qt o TO result.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, false)
end

#Create Half point model using data points from selection
def self.Selected_Half()
	Dir.chdir(@qhull_path)
	sel = Sketchup.active_model.selection
	return if (sel.empty?)
	array = []
	GetPoints(sel, array, Geom::Transformation.new, Geom::BoundingBox.new)
	file = File.new("input.txt", "w")
	file.puts("3 rbox " + array.length.to_s + " t")
	file.puts(array.length.to_s)
	array.each do |v|
		file.puts(v.x.to_f.to_s + " " + v.y.to_f.to_s + " " + v.z.to_f.to_s)
	end
	file.puts("")
	file.close()
	if not system("qconvex FQ FV n < input.txt | qhalf Qt o TO result.txt")
		UI.messagebox("QHull failed to generate valid results", MB_OK)
	end
	CommonResults(false, false)
end

#create SU geoemtry
def self.geometry(vertices, faces)
	if (vertices.length <3 or faces.length<1)
		UI.messagebox("QHull generated empty results", MB_OK)
	end

	if (Sketchup.version[0,1]=="7")
		Sketchup.active_model.start_operation("Qhull Geometry", true)
	else
		Sketchup.active_model.start_operation("Qhull Geometry")
	end
	
	entities = Sketchup.active_model.active_entities
	group = entities.add_group
	faces.each do |face_list|
		verts = []
		counter = 0
		face_list.each do |index|
				pt = vertices[index]
				invalid = false
				verts.each do |v|
					if (v==pt)
						invalid = true
					end
				end
				if not invalid
					verts << pt
				end
		end
		begin
			if (verts.length()>2)
				group.entities.add_face(verts)
			end
		rescue
			# probably non-coplanar, though that's not supposed to happen, it does
		end
	end
	
	Sketchup.active_model.commit_operation
end

end  #Qhull

Qhull.initialize

if( not file_loaded?("qhull.rb") )

	plugin_menu = UI.menu("Plugins").add_submenu("Qhull")
	voronoi_menu = plugin_menu.add_submenu("Voronoi")
	voronoi_menu.add_item("Small(10)"){Qhull.Voronoi(10)}
	voronoi_menu.add_item("Medium(50)"){Qhull.Voronoi(50)}
	voronoi_menu.add_item("Large(100)"){Qhull.Voronoi(100)}
	voronoi_menu.add_item("Huge(500)"){Qhull.Voronoi(500)}
	voronoi_menu.add_item("Selection (Sphere)"){Qhull.Selected_Voronoi(true, false)}
	voronoi_menu.add_item("Selection (BoundingBox)"){Qhull.Selected_Voronoi(false, true)}
	voronoi_menu.add_item("Selection (Free)"){Qhull.Selected_Voronoi(false, false)}
	convex_menu = plugin_menu.add_submenu("Convex")
	convex_menu.add_item("Small(10)"){Qhull.Convex(10)}
	convex_menu.add_item("Medium(50)"){Qhull.Convex(50)}
	convex_menu.add_item("Large(100)"){Qhull.Convex(100)}
	convex_menu.add_item("Huge(500)"){Qhull.Convex(500)}
	convex_menu.add_item("Selection"){Qhull.Selected_Convex}
	del_menu = plugin_menu.add_submenu("Delaunay")
	del_menu.add_item("Small(10)"){Qhull.Delaunay(10)}
	del_menu.add_item("Medium(50)"){Qhull.Delaunay(50)}
	del_menu.add_item("Large(100)"){Qhull.Delaunay(100)}
	del_menu.add_item("Huge(500)"){Qhull.Delaunay(500)}	
	del_menu.add_item("Selection"){Qhull.Selected_Delaunay}
	half_menu = plugin_menu.add_submenu("Half")
	half_menu.add_item("Small(10)"){Qhull.Half(10)}
	half_menu.add_item("Medium(50)"){Qhull.Half(50)}
	half_menu.add_item("Large(100)"){Qhull.Half(100)}
	half_menu.add_item("Huge(500)"){Qhull.Half(500)}
	half_menu.add_item("Selection"){Qhull.Selected_Half}
end

file_loaded("qhull.rb")

end # SU_Qhull